/* -----------------------------------------------------------------------------
 * d.cxx
 *
 *     D wrapper module.
 *
 * Author(s) : Andy Friesen
 *
 * Copyright (C) 2004
 * See the file LICENSE for information on usage and redistribution.
 * ----------------------------------------------------------------------------- */

char cvsroot_d_cxx[] = "$Header: /cvsroot/SWIG/Source/Modules/d.cxx,v 1.50 2003/12/23 21:45:38 cheetah Exp $";

#include <limits.h> // for INT_MAX
#include <ctype.h>
#include "swigmod.h"
#ifndef MACSWIG
#   include "swigconfig.h"
#endif

#include <cassert>
#include <string>
#include <vector>
#include <map>

namespace {

    template <typename T>
    void destroy(T* t) {
        Delete(t);
    }

    template<>
    void destroy<Wrapper>(Wrapper* w) {
        DelWrapper(w);
    }

    struct ScopedPtrBase { };

    template <class T>
    struct ScopedPtr : ScopedPtrBase {
        ScopedPtr(T* t)  : _data(t) { }
        ScopedPtr()      : _data(0) { }

        ScopedPtr(ScopedPtr& rhs) {
            (*this) = rhs;
        }

        virtual ~ScopedPtr() { 
            if (_data != 0) {
                destroy(_data);
            }
        }

        T* operator ->() const { return  _data; }
        T* get()         const { return  _data; }
     
        ScopedPtr& operator = (T* t) {
            destroy(_data);
            _data = t;
            return *this;
        }

        ScopedPtr& operator = (ScopedPtr& rhs) {
            (*this) = rhs.get();
            rhs = 0;
            return *this;
        }

    private:
        T* _data;
        //ScopedPtr(ScopedPtr&) { } // disallow copying
    };

    typedef ScopedPtr<File> ScopedFile;
    typedef ScopedPtr<String> ScopedString;

    template <typename T>
    ScopedPtr<T> makeAutoPtr(T* t) {
        return ScopedPtr<T>(t);
    }
}

// Evil little macro for automatically destroying a local when its scope dies.
// magic: Assigning a temporary to a const& extends the life of the temporary
// to that of the reference.
// NOTE: this is dumb and pointless.  Kill.
#define AUTO_HIDDEN(x, y) x ## y
#define AUTO_HIDE2(x, y) AUTO_HIDDEN(x, y)
#define AUTO(x) ScopedPtrBase const& AUTO_HIDE2(auto_secret, __LINE__) = makeAutoPtr(x);

/**
 * D language wrapper
 * TODO: deal with nested types.
 * FIXME: unnamed parameters in the source text cause incorrect output in
 * the output C++ code.
 */
class D : public Language {

    /// Context for functions: saves deduction effort.
    /// Stored in the node as its funcType property
    enum FuncType {
        ft_global       = (1 << 0), // is not a member of a class
        ft_static       = (1 << 1), // is a static member of a class
        ft_method       = (1 << 2), // is a member of a class
        ft_overload     = (1 << 3), // has overloads
        ft_ctor         = (1 << 4), // is a ctor
        ft_dtor         = (1 << 5),
        ft_attribute    = (1 << 6), // is an attribute
        ft_attribute_get= (1 << 7), // is an attribute get function
        ft_attribute_set= (1 << 8),

        // Composites for convenience
        ft_staticmethod = ft_static | ft_method,
    };

    virtual void main(int argc, char** argv) {
        SWIG_library_directory("d");
        Preprocessor_define("SWIGD 1", 0);
        SWIG_config_file("d.swg");
        SWIG_typemap_lang("d");

        ScopedString cname = NewStringf("test.cpp");
        ScopedString dimp  = NewStringf("test_extern.d");
        ScopedString dname = NewStringf("test.d");

        c_source = NewFile(cname.get(), "w");
        d_import = NewFile(dimp .get(), "w");
        d_source = NewFile(dname.get(), "w");

#if 0
        c_init   = NewString("");
#endif

        // SWIG uses these names to insert code in various places.
        // Link them up with actual file handles.
        Swig_register_filebyname("runtime", c_source.get());
        Swig_register_filebyname("wrapper", d_source.get());

        // We don't actually have a C header: 
        // just put it in the main C++ source file for now.
        Swig_register_filebyname("header", c_source.get());

        allow_overloading();
    }

    virtual int top(Node* n) {
        Language::top(n);
        return SWIG_OK;
    }

    //
    virtual int enumDeclaration(Node* n) {
        Printf(stderr, "enum %s\n", Getattr(n, "name"));

        Printf(d_source.get(), "enum %s {\n", Getattr(n, "name"));
        Language::enumDeclaration(n);
        Printf(d_source.get(), "}\n");
        return SWIG_OK;
    }

    virtual int enumvalueDeclaration(Node* n) {
        Printf(d_source.get(), "    %s = %s,\n", Getattr(n, "name"), Getattr(n, "enumvalue"));
        return SWIG_OK;
    }

    //
    virtual int globalfunctionHandler(Node* node) {
        // Already handled
        if (hasattr(node, "funcType")) {
            return Language::globalfunctionHandler(node);
        }

        int result = SWIG_ERROR;

        int flags = GetInt(node, "funcType");
        SetInt(node, "funcType", flags | ft_global);

        if (hasattr(node, "sym:overloaded")) {
            // overloads need proxies so overloading works correctly.
            // pass-through to generate proxies.
            Printf(stderr, "global overloaded function %s\n", Getattr(node, "name"));

            result = Language::globalfunctionHandler(node);
        } else {
            Printf(stderr, "Skipping non-overloaded global function %s: D needs no glue.\n", Getattr(node, "name"));
            result = SWIG_NOWRAP;
        }

        SetInt(node, "funcType", flags);
        return result;
    }

    virtual int functionHandler(Node* node) {
        int flags = GetInt(node, "funcType");
        const int oldFlags = flags;

        if (hasattr(node, "sym:overloaded")) {
            flags |= ft_overload;
        }
        if (getCurrentClass() != 0) {
            flags |= ft_method;
        }
        if (hasattr(node, "storage") && 
            Cmp(Getattr(node, "storage"), "static") == 0) 
        {
            flags |= ft_static;
        }

        if (!(flags & ft_method)) {
            // is this the right thing?
            flags |= ft_global;
        }

        SetInt(node, "funcType", flags);

        int result = SWIG_ERROR;
        if ((flags & ft_staticmethod) == ft_staticmethod) {
            result = staticmemberfunctionHandler(node);
        } else if (flags & ft_method) {
            result = memberfunctionHandler(node);
        } else if (flags & ft_global) {
            result = globalfunctionHandler(node);
        } else {
            assert(false);
            return SWIG_ERROR;
        }
        SetInt(node, "funcType", oldFlags);
        return result;
    }

    virtual int memberfunctionHandler(Node* n) {
        Printf(stderr, "\tmethod: %s\n", Getattr(n, "name"));

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_method);

        // HACK: don't double-mangle get/setters
        if (!(flags & ft_attribute)) {
            Swig_MethodToFunction(n, Getattr(getCurrentClass(), "name"), 0);
        }
        int result = functionWrapper(n);
        SetInt(n, "funcType", flags);
        return result;
    }

    virtual int staticmemberfunctionHandler(Node* n) {
        Printf(stderr, "\tstatic method %s\n", Getattr(n, "name"));

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_method | ft_static);
        
        // skip the default static method handler: it mangles names in a bad way
        int result = globalfunctionHandler(n);
        SetInt(n, "funcType", flags);

        return result;
    }

    virtual int callbackfunctionHandler(Node* n) {
        return NYI(n);
    }

    //
    virtual int globalvariableHandler(Node *n) {
        Printf(stderr, "\tglobal attribute %s\n", Getattr(n, "name"));
        return Language::globalvariableHandler(n);
    }

    virtual int membervariableHandler(Node *n) {
        Printf(stderr, "\tattribute %s\n", Getattr(n, "name"));
        // TODO: use memberin typemap

        int flags = GetInt(n, "funcType");

        if (1) {
            // emit setter
            Node* oldType = Getattr(n, "type");
            
            SetInt(n, "funcType", flags | ft_attribute | ft_attribute_get);
            Swig_MembergetToFunction(n, Getattr(getCurrentClass(), "name"), 0);
            functionHandler(n);
        }

        SetInt(n, "funcType", flags | ft_attribute | ft_attribute_set);
        Swig_MembersetToFunction(n, Getattr(getCurrentClass(), "name"), 0);
        functionHandler(n);
        //variableWrapper(n);
        SetInt(n, "funcType", flags);

        return SWIG_OK;
    }

    virtual int staticmembervariableHandler(Node *n) {
#if 1

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_static | ft_attribute);
        //int result = Language::variableWrapper(n);
        int result = Language::staticmembervariableHandler(n);
        dumpNode(n);
        SetInt(n, "funcType", flags);
        return result;

#else

        Printf(stderr, "\tstatic attribute %s\n", Getattr(n, "name"));

        int flags = GetInt(n, "funcType");

        // Emit getter
        SetInt(n, "funcType", flags | ft_attribute_get | ft_static);
        Swig_VargetToFunction(n);
        functionHandler(n);

        if (1) {
            // Emit setter
            SetInt(n, "funcType", flags | ft_attribute_set | ft_static);
            Swig_VarsetToFunction(n);
            functionHandler(n);
        }

        SetInt(n, "funcType", flags);

        return SWIG_OK;
#endif
    }

    //
    virtual int memberconstantHandler(Node* n) {
        Printf(stderr, "\tconstant attribute %s\n", Getattr(n, "name"));
        return Language::memberconstantHandler(n);
    }

    virtual int constructorHandler(Node* n) {
        Printf(stderr, "\tctor %s\n", Getattr(n, "name"));

        SetInt(n, "funcType", ft_ctor);
        int result = Language::constructorHandler(n);
        delattr(n, "funcType");
        return result;
    }

    virtual int copyconstructorHandler(Node* n) {
        Printf(stderr, "\tcopy ctor %s\n", Getattr(n, "name"));                

        SetInt(n, "funcType", ft_ctor);
        int result = Language::copyconstructorHandler(n);
        delattr(n, "funcType");
        return result;
    }

    virtual int destructorHandler(Node* n) {
        Printf(stderr, "\tdtor\n");

        SetInt(n, "funcType", ft_dtor);
        int result = Language::destructorHandler(n);
        delattr(n, "funcType");
        return result;
    }

    // Maybe still too complex.  Simplify?
    virtual int classHandler(Node* n) {
        Printf(stderr, "\nclass %s:\n", Getattr(n, "name"));
        
        String* className = Getattr(n, "name");
        String* baseName = NewString("");
        bool hasBase = hasattr(n, "bases");
        AUTO(baseName);

        if (hasBase) {
            List* baseList = Getattr(n, "bases");
            Iterator iter = First(baseList);
            if (iter.item) {
                baseName = Getattr(iter.item, "name");
            }
            if (Next(iter).item != 0) {
                // As good a warning as any.
                Swig_warning(WARN_JAVA_MULTIPLE_INHERITANCE, input_file, line_number,
                    "Warning for %s proxy: Base %s ignored. Multiple inheritance "
                    "is not supported in D.\n", className, Getattr(iter.item,"name")
                );
            }
        }

        // impl pointer, protected constructor (used by all public constructors and subclasses)
        if (!hasBase) {
            Printv(d_source.get(),
                "class ", className, " {\n",
                "\n",
                "private void* swig_impl;\n",
                "private bool  swig_own_ptr;\n",
                "protected this(void* ptr, bool own) {\n",
                "    swig_impl = ptr;\n",
                "    swig_own_ptr = own;\n",
                "}\n",
                "\n",
                NIL
            );
        } else {
            Printv(d_source.get(),
                "class ", className, " : ", baseName, " {\n",
                "\n",
                "private void* swig_impl;\n",
                "protected this(void* ptr, bool own) {\n",
                "    super(ptr, own);\n",
                "    swig_impl = ptr;\n",
                "}\n",
                "\n",
                NIL
            );
        }

        Language::classHandler(n);

        Printf(d_source.get(), "} // end class %s\n\n", className);

        return SWIG_OK;
    }

    //
    virtual int constantWrapper(Node* n) {
        String* varType = Getattr(n, "type");
        String* varName = Getattr(n, "name");
        String* varValue = quoteConstant(n);
        AUTO(varValue);

        Printf(d_source.get(), "const %s %s = %s;\n", varType, varName, varValue);

        return SWIG_OK;
    }

    virtual int functionWrapper(Node* node) {
        Setattr(node, "sym:name", mangleName(node));
        Language::functionWrapper(node);
        int result;

        int flags = GetInt(node, "funcType");

        // C glue
        result = writeExternCFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // D glue
        result = writeExternDFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // wrapper interface
        result = writeWrapper(node);
        return result;
    }

    /*virtual int variableWrapper(Node* node) {
        int flags = GetInt(node, "funcType");
        if (!(flags & ft_global)) {
            Swig_MembergetToFunction(node, Getattr(getCurrentClass(), "name"), 0);
            SetInt(node, "funcType", flags | ft_attribute_get);
            functionWrapper(node);
            Swig_MembersetToFunction(node, Getattr(getCurrentClass(), "name"), 0);
            SetInt(node, "funcType", flags | ft_attribute_set);
            functionWrapper(node);
            SetInt(node, "funcType", flags);
            return SWIG_OK;
        } else {
            Printf(stderr, "variable:\n");
            dumpNode(node);
            return Language::variableWrapper(node);
        }
    }*/

    int writeExternCFunc(Node* node) {
        String*    name  = Getattr(node,"sym:name");
        SwigType*  type  = Getattr(node,"type");
        ParmList*  parms = Getattr(node,"parms");

        String* returnType = SwigType_str(type, 0);
        AUTO(returnType);

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        emit_attach_parmmaps(parms, wrapper);
        Swig_typemap_attach_parms("CtoExtern", parms, wrapper);

        Printf(wrapper->def, "DllExport %s CDecl %s(%s) {", returnType, name, ParmList_str(parms));

        emit_args(type, parms, wrapper);

        emit_action(node, wrapper);

        if (SwigType_type(type) != T_VOID) {
            Printf(wrapper->code, "return result;");
        }

        Printf(wrapper->code, "}");
        Wrapper_print(wrapper, c_source.get());

        writeExternDFunc(node);

        return SWIG_OK;
    }

    int writeExternDFunc(Node* node) {
        String*    name  = Getattr(node,"sym:name");
        SwigType*  type  = Getattr(node,"type");
        ParmList*  parms = Getattr(node,"parms");

        String*    argList = NewString("");
        AUTO(argList);

        String* returnType = SwigType_str(type, 0);

        Swig_typemap_attach_parms("CtoD", parms, 0);

        String* mappedReturnType = Swig_typemap_lookup_new("CtoD", node, "", 0);

        int parmCount = emit_num_arguments(parms);

        // Construct argument list: transform each argument type
        // and make up a name.
        Parm* parm = parms;
        for (int i = 0; i < parmCount; i++) {
            String* mappedType = Getattr(parm, "tmap:CtoD");

            String* oldType = SwigType_str(Getattr(parm, "type"), 0);

            if (mappedType != 0) {
                Printv(argList, mappedType, NIL);
            } else {
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, name);
                Printv(argList, oldType, NIL);
            }

            Printf(argList, " arg%i", i + 1);

            if (i < parmCount - 1) {
                Printf(argList, ", ");
            }

            parm = nextSibling(parm);
        }

        if (mappedReturnType == 0) {
            mappedReturnType = returnType;
        }

        Printf(d_import.get(), "extern (C) %s %s(%s);\n", mappedReturnType, name, argList);

        return SWIG_OK;
    }

    int writeWrapper(Node* node) {
        // Now the D wrapper

        if (!hasattr(node, "funcType")) {
            Printf(stderr, "Internal error on function %s in D::writeWrapper.  Maybe?\n", Getattr(node, "name"));
            return SWIG_ERROR;
        }

        int flags = GetInt(node, "funcType");

        bool isAttribute = bool(flags & ft_attribute);
        bool isCtor = bool(flags & ft_ctor);
        bool isDtor = bool(flags & ft_dtor);
        bool isMethod = bool(flags & ft_method);
        bool isStatic = bool(flags & ft_static);
        bool isGlobal = bool(flags & ft_global);

        int result = SWIG_ERROR;
        if (isCtor) {
            result = writeWrapperCtor(node);

        } else if (isDtor) {
            result = writeWrapperDtor(node);

        } else if (isMethod || isStatic || isGlobal || isAttribute) {
            result = writeWrapperFunc(node, isGlobal, isStatic);

        } else {
            Printf(stderr, "Skipping %s %s: D needs no wrapper for pure C things.\n", 
                Getattr(node, "nodeType"), Getattr(node, "name")
            );

            result = SWIG_OK;
        }

        return result;
    }

    int writeWrapperFunc(Node* node, bool isGlobal, bool isStatic) {
        String*   name       = Getattr(node, "name");       // pretty name
        String*   wrapName   = Getattr(node, "sym:name");   // mangled name
        SwigType* returnType = Getattr(node, "type");       // return type
        ParmList* parms      = Getattr(node, "parms");
        bool hasThis         = !isGlobal && !isStatic;

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        String* mappedReturnType = Swig_typemap_lookup_new("CtoD", node, "", 0);
        if (mappedReturnType == 0) {
            mappedReturnType = returnType;
        }

        Swig_typemap_attach_parms("DtoC", parms, wrapper);

        String* argList = NewString("");
        AUTO(argList);

        Parm* parm = parms;
        int parmCount = emit_num_arguments(parms);

        if (hasThis && parm) {
            // skip the first argument if we have a 'this' pointer
            parm = nextSibling(parm);
            parmCount--;
        }

        for (int i = 0; i < parmCount; i++) {
            String* mappedType = Getattr(parm, "tmap:CtoD");

            if (mappedType == 0) {
                String* oldType = SwigType_str(Getattr(parm, "type"), 0);
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, wrapName);
                mappedType = oldType; // copy existing type verbatim. (it might work!!)
            }
    
            Printf(argList, "%s arg%i", mappedType, i + 1);

            if (i < parmCount - 1) {
                Printv(argList, ", ", NIL);
            }

            parm = nextSibling(parm);
        }

        if (isStatic) {
            Printf(wrapper->def, "static ");
        }

        Printf(wrapper->def, "%s %s(%s) {", mappedReturnType, name, argList);

        if (hasThis) {
            Printf(wrapper->code, "void* arg1 = swig_impl;\n");
        }

        if (SwigType_type(returnType) != T_VOID) {
            Printv(wrapper->code, "return ", NIL);
        }
        Printv(wrapper->code, Swig_cfunction_call(wrapName, parms), ";", NIL);

        Printf(wrapper->code, "}");

        Wrapper_print(wrapper, d_source.get());

        return SWIG_OK;
    }

    int writeWrapperCtor(Node* node) {
        ParmList* parms = Getattr(node, "parms");
        String* name = Getattr(node, "name");
        String* wrapName = Getattr(node, "sym:name");

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        Swig_typemap_attach_parms("DtoC", parms, wrapper);

        // argument names and types
        // ie double d1, double d2, ...
        String* argList = NewString("");
        AUTO(argList);

        // argument names, no types
        // ie d1, d2, ...
        String* callList = NewString("");

        int parmCount = emit_num_arguments(parms);
        Parm* iter = parms;
        for (int i = 0; i < parmCount; i++) {
            SwigType* parmType = Getattr(iter, "type");
            String* parmName = Getattr(iter, "name");

            String* mappedType = Getattr(iter, "tmap:DtoC");
            if (mappedType == 0) {
                String* oldType = SwigType_str(Getattr(iter, "type"), 0);
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, parmName);
                mappedType = oldType;
            }

            Printf(argList, "%s arg%i", mappedType, i + 1);
            Printf(callList, "arg%i", i + 1);

            if (i < parmCount - 1) {
                Printv(argList, ", ", NIL);
                Printv(callList, ", ", NIL);
            }

            iter = nextSibling(iter);
        }

        Printf(wrapper->def, "this(%s) {", argList);
        
        Printf(wrapper->code, "this(%s(%s), true);\n", wrapName, callList);
        Printf(wrapper->code, "}");

        //Printf(stderr, "CTOR\t%s\n", argList);
        Wrapper_print(wrapper, d_source.get());

        return SWIG_OK;
    }

    int writeWrapperDtor(Node* node) {
        // Base class defines protected swig_destroy() and ~this()
        // Subclasses just override swig_destroy()
        bool hasBase = hasattr(node, "bases");
        String* dtorName = Getattr(node, "sym:name");
        if (!hasBase) {
            // Root class: define swig_destroy method and actual destructor.
            Printf(d_source.get(),
                "protected void swig_destroy()\n"
                "out {\n"
                "    assert(swig_impl === null);\n"
                "}\n"
                "body {\n"
                "    if (swig_own_ptr && swig_impl !== null) {\n"
                "        %s(swig_impl);\n"
                "    }\n"
                "    swig_impl = null;\n"
                "}\n"
                "\n"
                "~this() {\n"
                "    swig_destroy();\n"
                "}\n"
                "\n",
                dtorName
            );
        } else {
            // No actual destructor on subclasses:
            // Just override swig_destroy() method
            Printf(d_source.get(), 
                "    protected override void swig_destroy() {\n"
                "        if (swig_own_ptr && swig_impl !== null) {\n"
                "            %s(swig_impl);\n"
                "        }\n"
                "        swig_impl = null;\n"
                "    }\n"
                "\n",
                dtorName
            );
        }
        return SWIG_OK;
    }

    /// Returns a mangled name for the function
    String* mangleName(Node* node) {
        int flags = GetInt(node, "funcType");

        String* name = Getattr(node, "name");

        String* result = NewString("SwigWrap_");

        if (flags & ft_ctor) {
            Printv(result, "new_", name, NIL);
        } else if (flags & ft_dtor) {
            Printv(result, "delete_", name, NIL);
        } else {
            // attribute, method

            if (flags & (ft_method | ft_static)) {
                // Append class name
                String* className = Getattr(getCurrentClass(), "name");
                Printv(result, className, "_", NIL);
            }

            // Append actual name
            Printv(result, name, NIL);
        }

        if (flags & ft_attribute_get) {
            Printv(result, "_get", NIL);
        } else if (flags & ft_attribute_set) {
            Printv(result, "_set", NIL);
        }

        if (flags & ft_overload) {
            // Append wart to disambiguate overloads
            Printv(result, Getattr(node, "sym:overname"), NIL);
        }

        Replaceall(result, "::", "_");
        return result;
    }

    void dumpNode(Node* node) {
            for (Iterator iter = First(node); iter.item; iter = Next(iter)) {
            if (DohIsString(iter.item)) {
                Printf(stderr, "'%s'\t'%s'\n", iter.key, iter.item);
            } else {
                Printf(stderr, "'%s'\t<%s>\n", iter.key, Getattr(iter.item, "nodeType"));
            }
        }
        Printf(stderr, "\n\n\n");
    }

    int NYI(Node* node) {
        Printf(stderr, "Not yet implemented: %s %s\n", Getattr(node, "nodeType"), Getattr(node, "name"));
        dumpNode(node);
        return SWIG_ERROR;
    }

    bool hasThis(Node* node) {
        if (!Getattr(node, "funcType")) {
            return false;
        }

        int i = GetInt(node, "funcType");
        if (i & (ft_static | ft_global)) {
            return false;
        } else {
            return true;
        }
    }

    bool hasattr(Node* node, String* key) {
        return Getattr(node, key) != 0;
    }

    void delattr(Node* node, String* key) {
        Setattr(node, key, 0);
    }

    // Puts the quotes on string/char literals
    // (SWIG strips them for some reason)
    String* quoteConstant(Node* value) {
        int valueType = SwigType_type(Getattr(value, "type"));
        if (valueType == T_STRING || valueType == T_CHAR) {
            char quote = valueType == T_STRING ? '\"' : '\'';
            String* oldValue = Getattr(value, "value");

            return NewStringf("%c%s%c", quote, oldValue, quote);
        } else {
            return Copy(Getattr(value, "value"));
        }
    }

    // attributes:

    /// Contains the implementation of the wrapped C API
    ScopedFile c_source;

    /// D source file that declares extern(C) functions for 
    /// the contents of c_source (above)
    ScopedFile d_import;

    /// Wraps the C API in a delicious, candy coated interface.
    ScopedFile d_source;

    /// Not used. (not useful?  Provide anyway for the sake of compatibility with other backends?)
#if 0
    ScopedString c_init;   // init code. (also goes into c_source, I think)
#endif

};   /* class D */

extern "C" Language *
swig_d(void) {
  return new D();
}

#if 0
const char *D::usage = reinterpret_cast<char*>("\
D Options (available with -d)\n\
     -package <name> - set name of the package to <name>\n\
\n");
#endif
